INPUT_CHANNELS = 3
INPUT_HEIGHT = 48
INPUT_WIDTH = 48
INPUT_SIZE = (INPUT_WIDTH, INPUT_HEIGHT)
INPUT_SHAPE = (INPUT_CHANNELS, INPUT_HEIGHT, INPUT_WIDTH)
INPUT_SIZE_DICT = dict(
    height=INPUT_HEIGHT,
    width=INPUT_WIDTH,
)

N_AUGMENTATIONS = 9
