from .adabins_depth import AdaBinsDepth
from .blip import BLIP
from .clip import CLIP
from .cloob import CLOOB
from .deep_image_prior import DeepImagePrior
from .glide_clip import GlideCLIP
from .guided_diffusion import GuidedDiffusion
from . import latent_diffusion
from .midas_depth import MidasDepth
from .monster_diffusion import MonsterDiffusion
from .owlvit import OWLViT
from .ruclip import RuCLIP
from .slip import SLIP
from .stable_diffusion import StableDiffusion
from .stylegan_xl import StyleGANXL
from .super_resolution import SuperResolution
from .velocity_diffusion import VelocityDiffusion
from .open_clip import OpenCLIP
from .simulacra_aesthetic import SimulacraAesthetic
