P_mean = -1.2
P_std = 1.2
sigma_data = 0.5
rho = 7
sigma_min = 1e-2  # 0.002
sigma_max = 80

# depends on dataset, see table 5 in https://arxiv.org/pdf/2206.00364.pdf
# ImageNet 64x64
S_tmin = 0.05
S_tmax = 50
S_churn = 80
S_noise = 1.003  # between 1.0 and 1.007
