model_urls = {
    "yfcc_2": "https://s3.eu-central-1.wasabisys.com/nextml-model-data/velocity-diffusion/yfcc_2.pth",
    "yfcc_1": "https://s3.eu-central-1.wasabisys.com/nextml-model-data/velocity-diffusion/yfcc_1.pth",
    "cc12m_1": "https://s3.eu-central-1.wasabisys.com/nextml-model-data/velocity-diffusion/cc12m_1.pth",
    "cc12m_1_cfg": "https://s3.eu-central-1.wasabisys.com/nextml-model-data/velocity-diffusion/cc12m_1_cfg.pth",
    "wikiart": "https://the-eye.eu/public/AI/models/v-diffusion/wikiart_256.pth",
}
